﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_GROUPVARINT_H_
#define INCLUDE_NN_NLIB_GROUPVARINT_H_

#include "nn/nlib/Config.h"
#ifdef NLIB_SIMD
#include "nn/nlib/simd/SimdInt.h"
#endif

NLIB_NAMESPACE_BEGIN

class GroupVarInt32 {
 public:
    static size_t GetEncodedSize(uint32_t v0, uint32_t v1,
                                 uint32_t v2, uint32_t v3) NLIB_NOEXCEPT {
        return (1 + 4 + 4 + 4 + 4) -
               (nlib_clz(v0 | 1) / 8) -
               (nlib_clz(v1 | 1) / 8) -
               (nlib_clz(v2 | 1) / 8) -
               (nlib_clz(v3 | 1) / 8);
    }
    static size_t GetEncodedSize(const uint32_t* p) NLIB_NOEXCEPT NLIB_NONNULL;
    static size_t GetNextGroupSize(const uint8_t* p) NLIB_NOEXCEPT NLIB_NONNULL;
    static uint8_t* Encode(uint8_t* p, uint32_t v0, uint32_t v1,
                           uint32_t v2, uint32_t v3) NLIB_NOEXCEPT NLIB_NONNULL;
    static uint8_t* Encode(uint8_t* p, const uint32_t* src) NLIB_NOEXCEPT NLIB_NONNULL;
    static const uint8_t* Decode(uint32_t* v0, uint32_t* v1,
                                 uint32_t* v2, uint32_t* v3,
                                 const uint8_t* p) NLIB_NOEXCEPT NLIB_NONNULL;
    static const uint8_t* Decode(uint32_t* dest, const uint8_t* p) NLIB_NOEXCEPT NLIB_NONNULL;
    static size_t GetMaxSize(size_t n) NLIB_NOEXCEPT {
        return ((n + 3) / 4) * (1 + 4 * sizeof(uint32_t));
    }
    static size_t GetTotalSize(const uint32_t* p, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;

 private:
#ifdef NLIB_SIMD
    NLIB_ALIGNAS(16) static NLIB_VIS_PUBLIC const uint8_t decodemasks_[256 * 16];
    NLIB_ALIGNAS(16) static NLIB_VIS_PUBLIC const uint8_t encodemasks_[256 * 16];
    static NLIB_VIS_PUBLIC const uint8_t count_[256];
    static NLIB_VIS_PUBLIC const uint8_t tagarray_[256];
#endif
};

inline size_t GroupVarInt32::GetEncodedSize(const uint32_t* p) NLIB_NOEXCEPT {
    return GetEncodedSize(p[0], p[1], p[2], p[3]);
}

inline size_t GroupVarInt32::GetNextGroupSize(const uint8_t* p) NLIB_NOEXCEPT {
#ifdef NLIB_SIMD
    return count_[*p];
#else
    uint8_t v = *p;
    return 5 + (v & 3) + ((v >> 2) & 3) + ((v >> 4) & 3) + ((v >> 6) & 3);
#endif
}

inline size_t GroupVarInt32::GetTotalSize(const uint32_t* p, size_t n) NLIB_NOEXCEPT {
    size_t rval = 0;
    for (; n > 4; n -= 4, p += 4) {
        rval += GetEncodedSize(p);
    }
    // The decoder must be able to read 1 + 16 bytes anywhere.
    // This is for _mm_loadu_si128 safety.
    rval += 1 + 16;
    return rval;
}

#ifdef NLIB_SIMD
inline uint8_t* GroupVarInt32::Encode(uint8_t* p, const uint32_t* src) NLIB_NOEXCEPT {
    simd::i128 val = simd::I128::LoadA4(src);
    int imask = simd::I128::MoveMask8(simd::I128::CmpEq8(val, simd::I128::SetZero()));
    uint8_t tag = tagarray_[imask & 0xFF] | (tagarray_[imask >> 8] << 4);

    *p = tag;
    simd::i128 mask = simd::I128::LoadA16(&encodemasks_[tag * 16]);
    simd::i128 r = simd::I128::Shuffle8(val, mask);
    simd::I128::StoreA1(p + 1, r);

    return p + count_[tag];
}

inline uint8_t* GroupVarInt32::Encode(uint8_t* p, uint32_t v0, uint32_t v1,
                                      uint32_t v2, uint32_t v3) NLIB_NOEXCEPT {
    simd::i128 val;
    val = simd::I128::SetUint32ToLane<0>(simd::I128::SetZero(), v0);
    val = simd::I128::SetUint32ToLane<1>(val, v1);
    val = simd::I128::SetUint32ToLane<2>(val, v2);
    val = simd::I128::SetUint32ToLane<3>(val, v3);
    int imask = simd::I128::MoveMask8(simd::I128::CmpEq8(val, simd::I128::SetZero()));
    uint8_t tag = tagarray_[imask & 0xFF] | (tagarray_[imask >> 8] << 4);

    *p = tag;
    simd::i128 mask = simd::I128::LoadA16(&encodemasks_[tag * 16]);
    simd::i128 r = simd::I128::Shuffle8(val, mask);
    simd::I128::StoreA1(p + 1, r);

    return p + count_[tag];
}
#else
inline uint8_t* GroupVarInt32::Encode(uint8_t* p, uint32_t v0, uint32_t v1,
                                      uint32_t v2, uint32_t v3) NLIB_NOEXCEPT {
    int v0k = 3 - (nlib_clz(v0 | 1) / 8);
    int v1k = 3 - (nlib_clz(v1 | 1) / 8);
    int v2k = 3 - (nlib_clz(v2 | 1) / 8);
    int v3k = 3 - (nlib_clz(v3 | 1) / 8);
    *p++ = static_cast<uint8_t>((v3k << 6) | (v2k << 4) | (v1k << 2) | v0k);

    // v0
    switch (v0k) {
    case 3:
        *(p + 3) = static_cast<uint8_t>(v0 >> 24);
        // fall through
    case 2:
        *(p + 2) = static_cast<uint8_t>((v0 >> 16) & 0xFF);
        // fall through
    case 1:
        *(p + 1) = static_cast<uint8_t>((v0 >> 8) & 0xFF);
        // fall through
    case 0:
        *p = static_cast<uint8_t>(v0);
        break;
    default:
        NLIB_ASSUME(0);
    }
    p += v0k + 1;

    // v1
    switch (v1k) {
    case 3:
        *(p + 3) = static_cast<uint8_t>(v1 >> 24);
        // fall through
    case 2:
        *(p + 2) = static_cast<uint8_t>((v1 >> 16) & 0xFF);
        // fall through
    case 1:
        *(p + 1) = static_cast<uint8_t>((v1 >> 8) & 0xFF);
        // fall through
    case 0:
        *p = static_cast<uint8_t>(v1);
        break;
    default:
        NLIB_ASSUME(0);
    }
    p += v1k + 1;

    // v2
    switch (v2k) {
    case 3:
        *(p + 3) = static_cast<uint8_t>(v2 >> 24);
        // fall through
    case 2:
        *(p + 2) = static_cast<uint8_t>((v2 >> 16) & 0xFF);
        // fall through
    case 1:
        *(p + 1) = static_cast<uint8_t>((v2 >> 8) & 0xFF);
        // fall through
    case 0:
        *p = static_cast<uint8_t>(v2);
        break;
    default:
        NLIB_ASSUME(0);
    }
    p += v2k + 1;

    // v3
    switch (v3k) {
    case 3:
        *(p + 3) = static_cast<uint8_t>(v3 >> 24);
        // fall through
    case 2:
        *(p + 2) = static_cast<uint8_t>((v3 >> 16) & 0xFF);
        // fall through
    case 1:
        *(p + 1) = static_cast<uint8_t>((v3 >> 8) & 0xFF);
        // fall through
    case 0:
        *p = static_cast<uint8_t>(v3);
        break;
    default:
        NLIB_ASSUME(0);
    }
    p += v3k + 1;

    return p;
}

inline uint8_t* GroupVarInt32::Encode(uint8_t* p, const uint32_t* src) NLIB_NOEXCEPT {
    return Encode(p, src[0], src[1], src[2], src[3]);
}
#endif

#ifdef NLIB_SIMD
inline const uint8_t* GroupVarInt32::Decode(uint32_t* dest, const uint8_t* p) NLIB_NOEXCEPT {
    uint8_t tag = *p;
    simd::i128 val = simd::I128::LoadA1(p + 1);
    simd::i128 mask = simd::I128::LoadA16(&decodemasks_[tag * 16]);
    simd::i128 r = simd::I128::Shuffle8(val, mask);
    simd::I128::StoreA4(dest, r);
    return p + count_[tag];
}

inline const uint8_t* GroupVarInt32::Decode(uint32_t* v0, uint32_t* v1,
                                            uint32_t* v2, uint32_t* v3,
                                            const uint8_t* p) NLIB_NOEXCEPT {
    uint8_t tag = *p;
    simd::i128 val = simd::I128::LoadA1(p + 1);
    simd::i128 mask = simd::I128::LoadA16(&decodemasks_[tag * 16]);
    simd::i128 r = simd::I128::Shuffle8(val, mask);

    *v0 = simd::I128::GetUint32FromLane<0>(r);
    *v1 = simd::I128::GetUint32FromLane<1>(r);
    *v2 = simd::I128::GetUint32FromLane<2>(r);
    *v3 = simd::I128::GetUint32FromLane<3>(r);
    return p + count_[tag];
}
#else
inline const uint8_t* GroupVarInt32::Decode(uint32_t* v0, uint32_t* v1,
                                            uint32_t* v2, uint32_t* v3,
                                            const uint8_t* p) NLIB_NOEXCEPT {
    uint8_t tag = *p;
    ++p;
    uint32_t tmp;

    // v0
    int v0k = tag & 3;
    tmp = 0;
    switch (v0k) {
    case 3:
        tmp |= *(p + 3) << 24;
        // fall through
    case 2:
        tmp |= *(p + 2) << 16;
        // fall through
    case 1:
        tmp |= *(p + 1) << 8;
        // fall through
    case 0:
        tmp |= *p;
        break;
    default:
        NLIB_ASSUME(0);
    }
    *v0 = tmp;
    p += v0k + 1;

    int v1k = (tag >> 2) & 3;
    tmp = 0;
    switch (v1k) {
    case 3:
        tmp |= *(p + 3) << 24;
        // fall through
    case 2:
        tmp |= *(p + 2) << 16;
        // fall through
    case 1:
        tmp |= *(p + 1) << 8;
        // fall through
    case 0:
        tmp |= *p;
        break;
    default:
        NLIB_ASSUME(0);
    }
    *v1 = tmp;
    p += v1k + 1;

    int v2k = (tag >> 4) & 3;
    tmp = 0;
    switch (v2k) {
    case 3:
        tmp |= *(p + 3) << 24;
        // fall through
    case 2:
        tmp |= *(p + 2) << 16;
        // fall through
    case 1:
        tmp |= *(p + 1) << 8;
        // fall through
    case 0:
        tmp |= *p;
        break;
    default:
        NLIB_ASSUME(0);
    }
    *v2 = tmp;
    p += v2k + 1;

    int v3k = (tag >> 6) & 3;
    tmp = 0;
    switch (v3k) {
    case 3:
        tmp |= *(p + 3) << 24;
        // fall through
    case 2:
        tmp |= *(p + 2) << 16;
        // fall through
    case 1:
        tmp |= *(p + 1) << 8;
        // fall through
    case 0:
        tmp |= *p;
        break;
    default:
        NLIB_ASSUME(0);
    }
    *v3 = tmp;
    p += v3k + 1;

    return p;
}

inline const uint8_t* GroupVarInt32::Decode(uint32_t* dest, const uint8_t* p) NLIB_NOEXCEPT {
    return Decode(&dest[0], &dest[1], &dest[2], &dest[3], p);
}
#endif

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_GROUPVARINT_H_
