﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_STRINGVIEW_H_
#define INCLUDE_NN_NLIB_STRINGVIEW_H_

#include <algorithm>
#include <iterator>

#include "nn/nlib/Config.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/Cstring.h"

NLIB_NAMESPACE_BEGIN

class NLIB_VIS_PUBLIC StringView {
 public:
    // types
    typedef char charT;
    typedef charT value_type;
    typedef const charT* pointer;
    typedef const charT* const_pointer;
    typedef const charT& reference;
    typedef const charT& const_reference;
    typedef const_pointer const_iterator;
    typedef const_iterator iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    typedef const_reverse_iterator reverse_iterator;
    typedef size_t size_type;
    typedef ptrdiff_t difference_type;
    static const size_type npos = size_type(-1);

    // [string.view.cons], construct/copy
    NLIB_CEXPR StringView() NLIB_NOEXCEPT : len_(0), base_("") {}
    // this ctor calls strlen(), use StringView(str, len) if you know the the length of the string.
    explicit StringView(const charT* str) NLIB_NOEXCEPT : len_(nlib_strlen(str)), base_(str) {
        NLIB_ASSERT(this->IsValid_());
    }
    StringView(const charT* str, size_type len) NLIB_NOEXCEPT : len_(len), base_(str) {
        NLIB_ASSERT(this->IsValid_());
    }

    // [string.view.iterators], iterators
    const_iterator begin() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return base_;
    }
    const_iterator end() const NLIB_NOEXCEPT { return begin() + len_; }
    const_iterator cbegin() const NLIB_NOEXCEPT { return begin(); }
    const_iterator cend() const NLIB_NOEXCEPT { return cbegin() + len_; }

    const_reverse_iterator rbegin() const NLIB_NOEXCEPT { return const_reverse_iterator(end()); }
    const_reverse_iterator rend() const NLIB_NOEXCEPT { return const_reverse_iterator(begin()); }
    const_reverse_iterator crbegin() const NLIB_NOEXCEPT { return const_reverse_iterator(end()); }
    const_reverse_iterator crend() const NLIB_NOEXCEPT { return const_reverse_iterator(begin()); }

    // [string.view.capacity], capacity
    size_type size() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return len_;
    }
    size_type length() const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        return len_;
    }
    size_type max_size() const NLIB_NOEXCEPT { return static_cast<size_type>(-1) / sizeof(charT); }
    bool empty() const NLIB_NOEXCEPT { return this->size() == 0; }

    // [string.view.access], element access
    const charT& operator[](size_type pos) const {
        NLIB_ASSERT(this->IsValid_());
        return base_[pos];
    }
    const charT& at(size_type pos) const {
        NLIB_ASSERT(this->IsValid_());
        return base_[pos];
    }
    const charT& front() const { return (*this)[0]; }
    const charT& back() const { return (*this)[this->length() - 1]; }
    const charT* data() const NLIB_NOEXCEPT { return base_; }

    // [string.view.modifiers], modifiers:
    void clear() NLIB_NOEXCEPT {
        len_ = 0;
        base_ = "";
    }
    void remove_prefix(size_type n) NLIB_NOEXCEPT {
        if (n >= len_) {
            base_ += len_;
            len_ = 0;
        } else {
            base_ += n;
            len_ -= n;
        }
    }
    void remove_suffix(size_type n) NLIB_NOEXCEPT {
        if (n >= len_) {
            len_ = 0;
        } else {
            len_ -= n;
        }
    }

    // [string.view.ops], string operations:
    StringView substr(size_type pos, size_type n = npos) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (pos >= len_) return StringView(base_ + len_, 0);
        const charT* new_base = base_ + pos;
        size_type r = static_cast<size_type>(len_ - pos);
        size_type newLen = r > n ? n : r;
        return StringView(new_base, newLen);
    }
    int compare(const StringView& s) const NLIB_NOEXCEPT;
    int compare(const charT* s) const NLIB_NOEXCEPT { return this->compare(StringView(s)); }
    bool starts_with(const StringView& s) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (s.length() > this->length()) return false;
        return nlib_memcmp(s.data(), base_, sizeof(*base_) * s.length()) == 0;
    }
    bool starts_with(charT c) const NLIB_NOEXCEPT { return front() == c; }
    bool starts_with(const charT* s) const NLIB_NOEXCEPT {
        return this->starts_with(StringView(s));
    }
    bool ends_with(const StringView& s) const NLIB_NOEXCEPT {
        NLIB_ASSERT(this->IsValid_());
        if (s.length() > len_) return false;
        return nlib_memcmp(s.data(), base_ + len_ - s.length(), sizeof(*base_) * s.length()) ==
               0;
    }
    bool ends_with(charT c) const NLIB_NOEXCEPT { return back() == c; }
    bool ends_with(const charT* s) const NLIB_NOEXCEPT { return this->ends_with(StringView(s)); }
    bool ToCstring(charT* str, size_type buf_size) const NLIB_NOEXCEPT;
    template <size_type N>
    bool ToCstring(charT (&str)[N]) const NLIB_NOEXCEPT {
        return ToCstring(str, N);
    }
#ifdef NLIB_CXX11_DEFAULT_TEMPLATE_ARGUMENT_FOR_FUNCTION_TEMPLATES
    template<class DUMMY = void>
#endif
    bool ToCstring(UniquePtr<charT[]>& str) const NLIB_NOEXCEPT {  // NOLINT
        // to avoid crossing the DLL boundary
        NLIB_ASSERT(this->IsValid_());
        str.reset(new (std::nothrow) charT[len_ + 1]);
        if (!str) return false;
        return ToCstring(str.get(), len_ + 1);
    }

    errno_t ToInteger(int32_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(int64_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(uint32_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(uint64_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(int8_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(int16_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(uint8_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToInteger(uint16_t* v, size_type* idx = NULL, int base = 10) const NLIB_NOEXCEPT;
    errno_t ToFloat(float* v, size_type* idx = NULL) const NLIB_NOEXCEPT;
    errno_t ToDouble(double* v, size_type* idx = NULL) const NLIB_NOEXCEPT;

    bool TrimLeft() NLIB_NOEXCEPT;
    bool TrimRight() NLIB_NOEXCEPT;
    void Trim() NLIB_NOEXCEPT {
        TrimLeft();
        TrimRight();
    }
    StringView GetLine() NLIB_NOEXCEPT;
    bool Proceed(const StringView& kwd) NLIB_NOEXCEPT {
        if (!this->starts_with(kwd)) return false;
        base_ += kwd.length();
        len_ -= kwd.length();
        return true;
    }
    bool Proceed(const charT* kwd) NLIB_NOEXCEPT { return this->Proceed(StringView(kwd)); }
    bool Proceed(charT ch) NLIB_NOEXCEPT;
    StringView GetName() NLIB_NOEXCEPT;
    /*
    size_type find(const StringView& s) const {
         StringView::iterator it = std::search(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type find(charT c) const;
    size_type find(const charT* s) const;
    size_type rfind(const StringView& s) const {
         StringView::iterator it = std::find_end(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type rfind(charT c) const;
    size_type rfind(const charT* s) const;
    size_type find_first_of(const StringView& s) const {
         StringView::iterator it = std::find_first_of(
             begin(), end(), s.begin(), s.end());
         return it == end() ? npos : it - begin();
    }
    size_type find_first_of(charT c) const;
    size_type find_first_of(const charT* s) const;
    size_type find_last_of(const StringView& s) const {
        StringView::reverse_iterator it = std::find_first_of(
            rbegin(), rend(), s.begin(), s.end());
        return it == rend() ? std::string::npos : it.base() - 1 - begin();
    }
    size_type find_last_of(charT c) const;
    size_type find_last_of(const charT* s) const;
    size_type find_first_not_of(const StringView& s) const;
    size_type find_first_not_of(charT c) const;
    size_type find_first_not_of(const charT* s) const;
    size_type find_last_not_of(const StringView& s) const;
    size_type find_last_not_of(charT c) const;
    size_type find_last_not_of(const charT* s) const;
    */

 private:
    bool IsValid_() const NLIB_NOEXCEPT {
        if (!base_) return false;
        return true;
    }
    size_type len_;
    const charT* base_;
};


inline errno_t StringView::ToInteger(int8_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    int32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 127 || tmp < -128) return ERANGE;
    *v = static_cast<int8_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(int16_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    int32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 32767 || tmp < -32768) return ERANGE;
    *v = static_cast<int16_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(uint8_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    uint32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 255) return ERANGE;
    *v = static_cast<uint8_t>(tmp);
    return 0;
}


inline errno_t StringView::ToInteger(uint16_t* v, size_type* idx, int base) const NLIB_NOEXCEPT {
    uint32_t tmp;
    errno_t e = ToInteger(&tmp, idx, base);
    if (e != 0) return e;
    if (tmp > 65535) return ERANGE;
    *v = static_cast<uint16_t>(tmp);
    return 0;
}

inline bool StringView::Proceed(charT ch) NLIB_NOEXCEPT {
    if (this->starts_with(ch)) {
        ++base_;
        --len_;
        return true;
    } else {
        return false;
    }
}

inline StringView StringView::GetName() NLIB_NOEXCEPT {
    // isalpha()[isalnum()]*
    NLIB_ASSERT(this->IsValid_());
    const charT* s = base_;
    const charT* s_end = base_ + len_;
    charT c = (s != s_end) ? *s : '\0';
    if (!nlib_isalpha(c) && c != '_') return *this;
    for (;;) {
        ++s;
        c = (s != s_end) ? *s : '\0';
        if (!nlib_isalnum(c) && c != '_') break;
    }
    size_type len = s - base_;
    StringView rval(base_, len);
    base_ = s;
    len_ = len_ - len;
    return rval;
}

// [string.view.comparison], non-member StringView comparison functions
inline bool operator==(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) == 0;
}
inline bool operator!=(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}
inline bool operator<(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) < 0;
}
inline bool operator>(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) > 0;
}
inline bool operator<=(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) <= 0;
}
inline bool operator>=(const StringView& lhs, const StringView& rhs) NLIB_NOEXCEPT {
    return lhs.compare(rhs) >= 0;
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_STRINGVIEW_H_
