﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TEXTREADER_H_
#define INCLUDE_NN_NLIB_TEXTREADER_H_

#include "nn/nlib/Config.h"

NLIB_NAMESPACE_BEGIN

class InputStream;

// code snippets:
//   TextReader reader;
//   if (reader.Init() != 0 || reader.Open(&stream) != 0) { error }
//   while ((c = reader.Read()) >= 0)
//     c is a codepoint
//   if (!reader) { stream error if !stream, otherwise UTF-8 may be malformed }
//   reader.Close();
//   (stream.Close();)
class NLIB_VIS_PUBLIC TextReader {
 public:
    TextReader() NLIB_NOEXCEPT;
    virtual ~TextReader() NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT;
    errno_t Open(InputStream* stream) NLIB_NOEXCEPT;
    int Read() NLIB_NOEXCEPT {
        if (cur_ == bufend_) {
            this->FillBuffer();
            if (cur_ == bufend_) return -1;
        }
        int rval = *reinterpret_cast<unsigned char*>(cur_);
        if (!(rval & 0x80)) {
            ++cur_;
            if (rval == 0x0A) {
                pos_of_line_ = pos_of_buf_ + (cur_ - &buf_[0]);
                ++line_;
            }
            return rval;
        }
        if (!utf32_cache_) ConstructUtf32Cache();
        cur_ += utf8_bytecount_;
        rval = static_cast<int>(utf32_cache_);
        utf32_cache_ = 0;
        return rval;
    }
    int Peek() NLIB_NOEXCEPT {
        if (cur_ == bufend_) {
            this->FillBuffer();
            if (cur_ == bufend_) return -1;
        }
        int rval = *reinterpret_cast<unsigned char*>(cur_);
        if (!(rval & 0x80)) return rval;
        if (!utf32_cache_) ConstructUtf32Cache();
        return static_cast<int>(utf32_cache_);
    }
    int SkipWs() NLIB_NOEXCEPT {
        // skip ' ' \t \n (\r)
        if (cur_ + 1 < bufend_) {
            unsigned char c = *reinterpret_cast<unsigned char*>(cur_);
            if (c > ' ') return 0;
            if (c == ' ') {
                ++cur_;
                c = *reinterpret_cast<unsigned char*>(cur_);
                if (c > ' ') {
                    utf32_cache_ = 0;
                    return 1;
                }
                return this->SkipWs_(1);
            }
        }
        return this->SkipWs_(0);
    }
    bool ReadUntil(size_t* len, char* buf, size_t n, char delim) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    bool ReadUntil(size_t* len, char (&buf)[N], char delim) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadUntil(len, &buf[0], N, delim);
    }
    template <class T>
    bool ReadUntil(size_t* len, char* buf, size_t n, T pred) NLIB_NOEXCEPT;
    template <class T, size_t N>
    bool ReadUntil(size_t* len, char (&buf)[N], T pred) NLIB_NOEXCEPT {  // NOLINT
        return this->ReadUntil(len, buf, N, pred);
    }
    size_t ReadDecimalString(char* buf, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    template <size_t N>
    size_t ReadDecimalString(char (&buf)[N]) NLIB_NOEXCEPT {
        return this->ReadDecimalString(buf, N);
    }
    bool Proceed(const char* str, size_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    bool Proceed(char c) NLIB_NOEXCEPT {
        NLIB_ASSERT(!(c & 0x80) && c != 0x0A);
        if (cur_ == bufend_) {
            this->FillBuffer();
            if (cur_ == bufend_) return false;
        }
        if (*cur_ != c) return false;
        ++cur_;
        return true;
    }
    bool ProceedEx(const char* str) NLIB_NOEXCEPT;
    bool Close() NLIB_NOEXCEPT;
    void SetError(errno_t e) const NLIB_NOEXCEPT {
        if (errno_ == 0) errno_ = e;
    }
    errno_t GetErrorValue() const NLIB_NOEXCEPT { return errno_; }
    InputStream* GetStream() NLIB_NOEXCEPT { return stream_; }
    int GetLine() const NLIB_NOEXCEPT { return line_; }
    int GetColumn() const NLIB_NOEXCEPT {
        return static_cast<int>((cur_ - &buf_[0]) + 1 + pos_of_buf_ - pos_of_line_);
    }
    NLIB_SAFE_BOOL(TextReader, GetErrorValue() == 0)

 protected:
    char* GetCur() NLIB_NOEXCEPT { return cur_; }
    char* GetBufEnd() NLIB_NOEXCEPT { return bufend_; }
    void SetBufEnd(char* p) NLIB_NOEXCEPT { bufend_ = p; }

    // checks UTF-8 validity, converts CRLF to LF, and convert CR to LF
    virtual void FillBuffer_() NLIB_NOEXCEPT;

 private:
    void FillBuffer() NLIB_NOEXCEPT { this->FillBuffer_(); }
    NLIB_VIS_HIDDEN void CountNewLine(const char* mcur, const char* mend) NLIB_NOEXCEPT;
    void ConstructUtf32Cache() NLIB_NOEXCEPT;
    int SkipWs_(int base) NLIB_NOEXCEPT;

 private:
    char buf_[512 + 3 + 1];
    nlib_utf32_t utf32_cache_;
    char* cur_;
    char* bufend_;
    size_t utf8_bytecount_;
    InputStream* stream_;
    mutable ErrnoT errno_;
    int line_;
    size_t pos_of_line_;
    size_t pos_of_buf_;

    NLIB_DISALLOW_COPY_AND_ASSIGN(TextReader);
};

template <class T>
bool TextReader::ReadUntil(size_t* len, char* buf, size_t n, T pred) NLIB_NOEXCEPT {
    // NOTICE: not NULL terminated
    if (!buf) {
        *len = 0;
        return false;
    }
    utf32_cache_ = 0;
    char* p = buf;
    char* pend = buf + n;
    for (;;) {
        if (cur_ == bufend_) {
            this->FillBuffer();
            if (cur_ == bufend_) {
                *len = p - buf;
                return false;
            }
        }
        // T::operator()(const char* ptr);
        while (p != pend && cur_ != bufend_) {
            if ((*reinterpret_cast<unsigned char*>(cur_) & 0xC0) != 0x80 &&
                pred((const char*)cur_)) {  // NOLINT
                *len = p - buf;
                return true;
            }
            *p = *cur_;
            ++p;
            ++cur_;
        }
        if (p == pend) {
            if (cur_ != bufend_ && (*reinterpret_cast<unsigned char*>(cur_) & 0xC0) == 0x80) {
                do {
                    --p;
                    --cur_;
                } while ((*reinterpret_cast<unsigned char*>(cur_) & 0xC0) == 0x80);
            }
            *len = p - buf;
            return false;
        }
    }
}

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_TEXTREADER_H_
