﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_URI_H_
#define INCLUDE_NN_NLIB_URI_H_

#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/Nlist.h"

NLIB_NAMESPACE_BEGIN

class InputStream;
class OutputStream;
class MemoryOutputStream;

class NLIB_VIS_PUBLIC Uri NLIB_FINAL {
 public:
    // reserved    = gen-delims / sub-delims
    // gen - delims = ":" / "/" / "?" / "#" / "[" / "]" / "@"
    // sub - delims = "!" / "$" / "&" / "'" / "(" / ")" / "*" / "+" / "," / ";" / "="
    static bool IsReserved(int c) NLIB_NOEXCEPT;
    // unreserved  = ALPHA / DIGIT / "-" / "." / "_" / "~"
    static bool IsUnreserved(int c) NLIB_NOEXCEPT;
    // 0-9 / A-F / a-f
    static bool IsHexDigit(int c) NLIB_NOEXCEPT;
    static errno_t DecodeUriComponent(size_t* written, char* buf, size_t n, const char* first,
                                      const char* last) NLIB_NOEXCEPT;
    static errno_t DecodeUriComponent(size_t* written, char* buf, size_t n,
                                      const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, n, s, s + nlib_strlen(s));
    }
    static errno_t EncodeUriComponent(size_t* written, char* buf, size_t n, const char* first,
                                      const char* last, bool fragment_mode) NLIB_NOEXCEPT;
    static errno_t EncodeUriComponent(size_t* written, char* buf, size_t n, const char* s,
                                      bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, n, s, s + nlib_strlen(s), fragment_mode);
    }
    static errno_t DecodePath(size_t* written, char* buf, size_t n, const char* s) NLIB_NOEXCEPT;
    static errno_t EncodePath(size_t* written, char* buf, size_t n, const char* s) NLIB_NOEXCEPT;

    template <size_t N>
    static errno_t DecodeUriComponent(size_t* written, char (&buf)[N],
                                      const char* s) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, N, s);
    }
    template <size_t N>
    static errno_t DecodeUriComponent(size_t* written, char (&buf)[N], const char* first,
                                      const char* last) NLIB_NOEXCEPT {
        return DecodeUriComponent(written, buf, N, first, last);
    }
    template <size_t N>
    static errno_t EncodeUriComponent(size_t* written, char (&buf)[N], const char* s,
                                      bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, N, s, fragment_mode);
    }
    template <size_t N>
    static errno_t EncodeUriComponent(size_t* written, char (&buf)[N], const char* first,
                                      const char* last, bool fragment_mode) NLIB_NOEXCEPT {
        return EncodeUriComponent(written, buf, N, first, last, fragment_mode);
    }
    template <size_t N>
    static errno_t DecodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return DecodePath(written, buf, N, s);
    }
    template <size_t N>
    static errno_t EncodePath(size_t* written, char (&buf)[N], const char* s) NLIB_NOEXCEPT {
        return EncodePath(written, buf, N, s);
    }

 public:
    Uri() NLIB_NOEXCEPT : prv_(NULL) {}
    ~Uri() NLIB_NOEXCEPT;
    bool Parse(const char* str) NLIB_NOEXCEPT;
    // for example:
    //   http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag
    // scheme = "http"
    // userinfo = "foobar:pass"
    // host = "www.example.com"
    // port = "80"
    // path = "/my/dir/index.html"
    // query = "var=value"
    // fragment = "frag"
    bool SetUri(const char* scheme, const char* userinfo, const char* host,
                const char* port, const char* path, const char* query,
                const char* fragment) NLIB_NOEXCEPT;
    bool ComposeString(char* buf, size_t size) const NLIB_NOEXCEPT;
    template <size_t N>
    bool ComposeString(char (&buf)[N]) const NLIB_NOEXCEPT {
        return this->ComposeString(buf, N);
    }

    // AddBaseUri("g/", "http://a/b/c/d;p?q") -> http://a/b/c/g/
    // AddBaseUri("/g", "http://a/b/c/d;p?q") -> http://a/g
    // AddBaseUri("//g", "http://a/b/c/d;p?q") -> http://g
    bool AddBaseUri(const Uri& relative, const Uri& base) NLIB_NOEXCEPT;
    // "http" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetScheme() const NLIB_NOEXCEPT;
    // "foobar:pass" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetUserInfo() const NLIB_NOEXCEPT;
    // "www.example.com" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetHost() const NLIB_NOEXCEPT;
    // "80" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPort() const NLIB_NOEXCEPT;
    // "/my/dir/index.html" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetPath() const NLIB_NOEXCEPT;
    // "var=value" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetQuery() const NLIB_NOEXCEPT;
    // "frag" from "http://foobar:pass@www.example.com:80/my/dir/index.html?var=value#frag"
    const char* GetFragment() const NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;

 private:
    NLIB_VIS_HIDDEN bool SetScheme_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetUserInfo_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetHost_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetPort_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetPath_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetQuery_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool SetFragment_(const char* first, const char* last) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool MergePath(const char* base, const char* relative) NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN bool RemoveDotSegments(const char* path) NLIB_NOEXCEPT;

 private:
    struct UriPrivate;
    mutable UriPrivate* prv_;
};

NLIB_NAMESPACE_END

#endif  // INCLUDE_NN_NLIB_URI_H_
