﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_SUCCINCT_SBV_H_
#define INCLUDE_NN_NLIB_SUCCINCT_SBV_H_

#include <string.h>
#include <utility>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/SmartBitmap.h"
#include "nn/nlib/BinaryReader.h"
#include "nn/nlib/BinaryWriter.h"
#include "nn/nlib/TypeTraits.h"

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace succinct {

namespace detail {
static const unsigned int BLK_SIZE = 32;
}  // namespace detail

class NLIB_VIS_PUBLIC Set NLIB_FINAL {
 public:
    typedef uint32_t IdxType;
    Set() NLIB_NOEXCEPT : prv_(NULL) {}
    ~Set() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_1(Set, prv_);
    void swap(Set& rhs) NLIB_NOEXCEPT {
        SetPrivate* tmp = rhs.prv_;
        rhs.prv_ = prv_;
        prv_ = tmp;
    }
    bool Init(uint32_t bv_size) NLIB_NOEXCEPT;
    bool TurnOn(uint32_t idx) NLIB_NOEXCEPT;
    bool TurnOff(uint32_t idx) NLIB_NOEXCEPT;
    bool Build() NLIB_NOEXCEPT;
    bool Has(uint32_t idx) const NLIB_NOEXCEPT;
    uint32_t Rank1(uint32_t idx) const NLIB_NOEXCEPT;
    uint32_t Rank0(uint32_t idx) const NLIB_NOEXCEPT {
        uint32_t rank1 = this->Rank1(idx);
        return idx + 1 - rank1;
    }
    int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT;
    int32_t Select0(uint32_t nth) const NLIB_NOEXCEPT;
    size_t MemSize() const NLIB_NOEXCEPT;
    uint32_t GetBitVectorSize() const NLIB_NOEXCEPT;
    const uint32_t* GetBitVector() const NLIB_NOEXCEPT;
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT NLIB_NONNULL;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT NLIB_NONNULL;

 private:
    struct SetPrivate;
    mutable SetPrivate* prv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Set);
};

class NLIB_VIS_PUBLIC Sbv NLIB_FINAL {
 public:
    typedef uint32_t IdxType;
    Sbv() NLIB_NOEXCEPT : prv_(NULL) {}
    ~Sbv() NLIB_NOEXCEPT;
    NLIB_MOVE_MEMBER_HELPER_2(Sbv, prv_, set_);
    void swap(Sbv& rhs) NLIB_NOEXCEPT {
        using std::swap;
        set_.swap(rhs.set_);
        SbvPrivate* tmp = rhs.prv_;
        rhs.prv_ = prv_;
        prv_ = tmp;
    }
    bool Init(uint32_t bv_size) NLIB_NOEXCEPT;
    bool TurnOn(uint32_t idx) NLIB_NOEXCEPT { return set_.TurnOn(idx); }
    bool TurnOff(uint32_t idx) NLIB_NOEXCEPT { return set_.TurnOff(idx); }
    bool Build() NLIB_NOEXCEPT;
    bool Has(uint32_t idx) const NLIB_NOEXCEPT { return set_.Has(idx); }
    uint32_t Rank1(uint32_t idx) const NLIB_NOEXCEPT { return set_.Rank1(idx); }
    uint32_t Rank0(uint32_t idx) const NLIB_NOEXCEPT { return set_.Rank0(idx); }
    int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT;
    int32_t Select0(uint32_t nth) const NLIB_NOEXCEPT { return set_.Select0(nth); }
    size_t MemSize() const NLIB_NOEXCEPT;
    uint32_t GetBitVectorSize() const NLIB_NOEXCEPT { return set_.GetBitVectorSize(); }
    const uint32_t* GetBitVector() const NLIB_NOEXCEPT { return set_.GetBitVector(); }
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    struct SbvPrivate;
    SbvPrivate* prv_;
    Set set_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Sbv);
};

class NLIB_VIS_PUBLIC SparseSet NLIB_FINAL {
 public:
    // index is 64-bit and # of on-bits is 32-bit.
    // In other words, rank is 64-bit index, and select is 32-bit index
    typedef uint64_t IdxType;
    SparseSet() NLIB_NOEXCEPT : bitvector_size_(0),
                                lowest_idx_(0xFFFFFFFF),
                                highest_idx_(0),
                                num_on_bits_(0),
                                width_(0),
                                blen_(0),
                                b_(NULL) {}
    ~SparseSet() NLIB_NOEXCEPT;
#ifdef NLIB_CXX11_RVALUE_REFERENCES
    SparseSet(SparseSet&& rhs) NLIB_NOEXCEPT;
    SparseSet& operator=(SparseSet&& rhs) NLIB_NOEXCEPT;
#endif
    bool Init(uint64_t bv_size) NLIB_NOEXCEPT;
    bool TurnOn(uint64_t idx) NLIB_NOEXCEPT;
    bool Build() NLIB_NOEXCEPT;
    bool Has(uint64_t idx, uint32_t* rank) const NLIB_NOEXCEPT {
        // called after Build()
        uint32_t rank1 = this->Rank1(idx);
        if (rank) *rank = rank1;
        if (rank1 == 0) return false;
        int64_t select1 = this->Select1Ex(rank1 - 1);
        NLIB_ASSERT(select1 >= 0);
        // if (select1 < 0) return false;  // not possible
        return idx == static_cast<uint64_t>(select1);
    }
    bool Has(uint64_t idx) const NLIB_NOEXCEPT {
        return this->Has(idx, NULL);
    }
    uint32_t Rank1(uint64_t idx) const NLIB_NOEXCEPT;
    uint32_t Rank0(uint64_t idx) const NLIB_NOEXCEPT {
        uint32_t rank1 = this->Rank1(idx);
        return static_cast<uint32_t>(idx + 1 - rank1);
    }
    int32_t Select1(uint32_t nth) const NLIB_NOEXCEPT;
    int64_t Select1Ex(uint32_t nth) const NLIB_NOEXCEPT;
    size_t MemSize() const NLIB_NOEXCEPT;
    uint64_t GetBitVectorSize() const NLIB_NOEXCEPT {
        return bitvector_size_;
    }
    void Reset() NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;
    void swap(SparseSet& rhs) NLIB_NOEXCEPT;  // NOLINT

 private:
    uint32_t GetBits(uint64_t pos, size_t num) const NLIB_NOEXCEPT {
        NLIB_ASSERT(num <= 32);
        uint64_t idx64 = pos / detail::BLK_SIZE;
        NLIB_ASSERT(idx64 <= SIZE_MAX);
        size_t idx = static_cast<size_t>(idx64);
        size_t ofs = static_cast<size_t>(pos % detail::BLK_SIZE);
        if (ofs + num > 32) {
            uint64_t b = b_[idx + 1];
            b <<= 32;
            b += b_[idx];
            return static_cast<uint32_t>((b >> ofs) & ((1 << num) - 1));
        } else {
            return (b_[idx] >> ofs) & ((1 << num) - 1);
        }
    }

 private:
    uint64_t bitvector_size_;
    uint64_t lowest_idx_;
    uint64_t highest_idx_;
    uint32_t num_on_bits_;
    uint32_t width_;
    uint32_t blen_;
    uint32_t* b_;
    Sbv sbv_;  // Set sbv_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(SparseSet);
};

class NLIB_VIS_PUBLIC CompressedArray NLIB_FINAL {
    static const size_t UNIT_SIZE = 64;

 public:
    CompressedArray() NLIB_NOEXCEPT
        : header_(NULL), data_(NULL), header_size_(0), header_capacity_(0),
          data_size_(0), data_capacity_(0), workidx_(0) {}
    ~CompressedArray() NLIB_NOEXCEPT;
    CompressedArray(CompressedArray& rhs, move_tag) NLIB_NOEXCEPT;  // NOLINT
    CompressedArray& assign(CompressedArray &rhs, move_tag) NLIB_NOEXCEPT;  // NOLINT
#ifdef NLIB_CXX11_RVALUE_REFERENCES
    CompressedArray(CompressedArray&& rhs) NLIB_NOEXCEPT;
    CompressedArray& operator=(CompressedArray&& rhs) NLIB_NOEXCEPT;
#endif
    // NLIB_MOVE_MEMBER_HELPER_3(CompressedArray, header_, data_, workidx_);
    void swap(CompressedArray& rhs) NLIB_NOEXCEPT;
    bool PushBack(uint32_t x) NLIB_NOEXCEPT {
        work_[workidx_++] = x;
        if (workidx_ == UNIT_SIZE) return this->Build();
        return true;
    }
    uint32_t operator[](size_t idx) const NLIB_NOEXCEPT;
    size_t Size() const NLIB_NOEXCEPT {
        return workidx_ + header_capacity_ * UNIT_SIZE / 2;
    }
    void Reset() NLIB_NOEXCEPT;
    size_t MemSize() const NLIB_NOEXCEPT;
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT;
    bool Import(BinaryReader* r) NLIB_NOEXCEPT;

 private:
    bool Build() NLIB_NOEXCEPT;
    // Header:
    //  baseidx: index of data_(lower 27bits)
    //  k: bit width of diff data(upper 5bits)
    //  base: base value(uint32_t)

    // Data:
    //  UNIT_SIZE * kbit: diff data
    uint32_t* header_;
    uint32_t* data_;
    uint32_t header_size_;
    uint32_t header_capacity_;
    uint32_t data_size_;
    uint32_t data_capacity_;
    uint8_t workidx_;
    uint32_t work_[UNIT_SIZE];
    NLIB_DISALLOW_COPY_AND_ASSIGN(CompressedArray);
};

class Map NLIB_FINAL {
    typedef CompressedArray ArrayType;
    typedef SparseSet SetType;

 public:
    typedef SetType::IdxType IdxType;
    typedef std::pair<bool, uint32_t> FindType;
    Map() NLIB_NOEXCEPT : set_(), value_() {}
    ~Map() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_2(Map, set_, value_);
    void swap(Map& rhs) NLIB_NOEXCEPT {
        set_.swap(rhs.set_);
        value_.swap(rhs.value_);
    }
    bool Init(IdxType bv_size) NLIB_NOEXCEPT {
        if (value_.Size() != 0) return false;
        return set_.Init(bv_size);
    }
    bool TurnOn(IdxType idx, uint32_t data) NLIB_NOEXCEPT {
        if (!set_.TurnOn(idx)) return false;
        value_.PushBack(data);
        return true;
    }
    bool Build() NLIB_NOEXCEPT { return set_.Build(); }
    FindType Find(IdxType idx) NLIB_NOEXCEPT {
        if (set_.Has(idx))
            return std::make_pair(true, value_[set_.Rank1(idx) - 1]);
        else
            return std::make_pair(false, 0);
    }
    const FindType Find(IdxType idx) const NLIB_NOEXCEPT {
        if (set_.Has(idx))
            return std::make_pair(true, value_[set_.Rank1(idx) - 1]);
        else
            return std::make_pair(false, 0);
    }
    size_t MemSize() const NLIB_NOEXCEPT {
        size_t value_size = value_.MemSize();
        size_t set_size = set_.MemSize();
        return value_size + set_size;
    }
    const SetType& GetKeys() const NLIB_NOEXCEPT { return set_; }
    const ArrayType& GetValues() const NLIB_NOEXCEPT { return value_; }
    void Reset() NLIB_NOEXCEPT {
        value_.Reset();
        set_.Reset();
    }
    bool Export(BinaryWriter* w) const NLIB_NOEXCEPT {
        if (!value_.Export(w)) return false;
        if (!set_.Export(w)) return false;
        return true;
    }
    bool Import(BinaryReader* r) NLIB_NOEXCEPT {
        if (!value_.Import(r) || !set_.Import(r)) {
            this->Reset();
            return false;
        }
        return true;
    }

 private:
    SetType set_;
    ArrayType value_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Map);
};

}  // namespace succinct
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Set)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Sbv)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::SparseSet)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::CompressedArray)
NLIB_DEFINE_STD_SWAP(::nlib_ns::succinct::Map)

#if defined(_MSC_VER) && defined(nx_succinct_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_SUCCINCT_SBV_H_
