﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_TESTING_NTEST_H_
#define INCLUDE_NN_NLIB_TESTING_NTEST_H_

#include <string.h>
#include <new>

#include "nn/nlib/Config.h"
#include "nn/nlib/testing/Printer.h"
#include "nn/nlib/SimpleSingleton.h"
#include "nn/nlib/Nflags.h"

#if defined(_MSC_VER) && defined(nx_testing_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

// NOTE:
// The implementation does not need std::iostream(ostream).
NLIB_NAMESPACE_BEGIN
namespace testing {

NLIB_VIS_PUBLIC void InitNintendoTest(int* argc, char** argv);
NLIB_VIS_PUBLIC void InitNintendoTest(int* argc, wchar_t** argv);

class NLIB_VIS_PUBLIC Environment {
 public:
    virtual ~Environment() {}
    virtual void SetUp() {}
    virtual void TearDown() {}
};

NLIB_VIS_PUBLIC Environment* AddGlobalTestEnvironment(Environment* env);

// The current test result must be accessible globally.
struct NLIB_VIS_PUBLIC TestResult {
    bool HasFatalFaiulre() const { return is_fatal_error; }  // gtest
    bool HasNonfatalFailure() const { return is_non_fatal_error; }  // gtest

    bool is_disabled;
    bool is_done;
    bool is_fatal_error;
    bool is_non_fatal_error;

    // NOTE:
    // NO CONSTRUCTOR
    // The values are set by TestInfo::AddTestInfo() before main function.
    // If there is a constuctor, the constructor may be invoked
    // after TestInfo::AddTestInfo() executed.
};

class Test;
class NLIB_VIS_PUBLIC TestFactoryBase {
 public:
    virtual ~TestFactoryBase() {}
    virtual Test* Create() = 0;

 protected:
    TestFactoryBase() {}

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(TestFactoryBase);
};

template <class TestClass>
class TestFactory : public TestFactoryBase {
 public:
    TestFactory() {}
    virtual Test* Create() NLIB_OVERRIDE {
        NLIB_STATIC_ASSERT(sizeof(TestClass));
        return new TestClass;
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(TestFactory);
};

class NLIB_VIS_PUBLIC TestInfo NLIB_FINAL {
 public:
    TestInfo();
    ~TestInfo();
    const char* test_case_name() const { return testcase_name_; }  // gtest
    const char* name() const { return testname_; }                // gtest
    const TestResult* result() const { return &result_; }         // gtest
    static int Run();

 public:
    typedef void (*SetUpTestCaseFunc)();
    typedef void (*TearDownTestCaseFunc)();
    template <class TestClass>
    static TestInfo* AddTestInfo(const char* test_case, const char* test_name, int index,
                                 SetUpTestCaseFunc setup_func, TearDownTestCaseFunc teardown_func);

 private:
    static const int kMaxTestList = 4096;
    static TestInfo testlist_[kMaxTestList];
    static int testlist_idx_;

 private:
    bool RunTest();
    bool TestCaseNameLess(const TestInfo& rhs) const {
        int tmp = strcmp(testcase_name_, rhs.testcase_name_);
        if (tmp < 0) return true;
        if (tmp > 0) return false;
        return index_ < rhs.index_;
    }
    bool TestCaseNameEq(const TestInfo& rhs) const {
        return index_ == rhs.index_ &&
               strcmp(testcase_name_, rhs.testcase_name_) == 0;
    }
    bool IsTypedTest() const { return index_ >= 0; }
    int GetTypedTestVariant() const { return index_ < 0 ? 0 : index_; }
    struct TestSortPred {
        bool operator()(TestInfo* lhs, TestInfo* rhs) {
            return lhs->TestCaseNameLess(*rhs);
        }
    };

 private:
    static TestInfo& AddTestInfo_(const char* test_case, const char* test_name, int index,
                                  SetUpTestCaseFunc setup_func,
                                  TearDownTestCaseFunc teardown_func);

 private:
    friend class FollowMessageTerminator;
    const char* testcase_name_;
    const char* testname_;
    int index_;  // -1 if not TYPED_TEST_CASE
    TestFactoryBase* factory_;
    void (*setup_testcase_)();
    void (*teardown_testcase_)();
    TestResult result_;
    double testfactory_space_[2];
    NLIB_DISALLOW_COPY_AND_ASSIGN(TestInfo);
};

template <class TestClass>
TestInfo* TestInfo::AddTestInfo(const char* test_case, const char* test_name, int index,
                                SetUpTestCaseFunc setup_func, TearDownTestCaseFunc teardown_func) {
    // NOTE:
    // This function is executed before main function to register the tests.
    // DYNAMIC MEMORY ALLOCATION CANNOT BE PERFORMED in this function.
    // Some environments may initialize its heap in main function.
    TestInfo& info = AddTestInfo_(test_case, test_name, index, setup_func, teardown_func);
    void* p = &info.testfactory_space_[0];
    new (p) TestFactory<TestClass>();

    // NOTE:
    // We use type 'char*' to avoid
    // "warning: dereferencing type-punned pointer will break strict-aliasing rules"
    // on GCC. This is better than using -fno-strict-aliasing.
    char* tmpptr = reinterpret_cast<char*>(&info.testfactory_space_[0]);
    info.factory_ = reinterpret_cast<TestFactoryBase*>(tmpptr);

    return &info;
}

class NLIB_VIS_PUBLIC UnitTest NLIB_FINAL {
    // This code does nothing. it is only for source code compatibility with gtest.
 public:
    static UnitTest* GetInstance();

    // const char* original_working_dir() const;
    // const TestCase* current_test_case() const;

    const TestInfo* current_test_info() const { return current_; }

    // int random_seed() const;

    /*
    int successful_test_case_count() const;
    int failed_test_case_count() const;
    int total_test_case_count() const;
    int test_case_to_run_count() const;
    int successful_test_count() const;
    int failed_test_count() const;
    int disabled_test_count() const;
    int total_test_count() const;
    int test_to_run_count() const;
    int32_t elapsed_time() const;
    bool Passed() const;
    bool Failed() const;
    */

 private:
    TestInfo* current_;

 private:
    friend class TestInfo;
    friend class SimpleSingleton<UnitTest>;
    friend class FollowMessageTerminator;
    UnitTest() : current_(NULL) {}
    virtual ~UnitTest() {}

    NLIB_DISALLOW_COPY_AND_ASSIGN(UnitTest);
};

class NLIB_VIS_PUBLIC Test {
 public:
    static void SetUpTestCase() {}
    static void TearDownTestCase() {}

    Test();
    virtual ~Test();
    void Run();
    void RecordProperty(const char* key, const char* value);
    void RecordProperty(const char* key, int value);

 protected:
    virtual void SetUp() {}
    virtual void TearDown() {}
    virtual void TestBody() = 0;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Test);
};

class NLIB_VIS_PUBLIC AssertionResult NLIB_FINAL {
 public:
    explicit AssertionResult(bool result) : is_success_(result), builder_() {}
    ~AssertionResult() {}
    AssertionResult(const AssertionResult& rhs)
        : is_success_(rhs.is_success_), builder_() {
        // NOTE: avoid copying...
        builder_.swap(const_cast<AssertionResult&>(rhs).builder_);
    }
    operator bool() const { return is_success_; }
    AssertionResult operator!() const { return AssertionResult(!is_success_); }
    const char* c_str() const { return builder_.c_str(); }

    template <class T>
    AssertionResult& operator<<(const T& rhs) {
        builder_ << rhs;
        return *this;
    }
    AssertionResult& operator<<(AssertionResult& rhs) {  // NOLINT
        is_success_ = rhs.is_success_;
        builder_.swap(rhs.builder_);
        return *this;
    }

 private:
    bool is_success_;
    detail_func::DummyStream builder_;
    AssertionResult& operator=(const AssertionResult&);  // cannot assign
};

class NLIB_VIS_PUBLIC ScopedTrace NLIB_FINAL {
 public:
    static void Print();
    ScopedTrace(AssertionResult& msg, const char* file, int line);  // NOLINT
    ~ScopedTrace();

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(ScopedTrace);
};

class NLIB_VIS_PUBLIC FollowMessageTerminator NLIB_FINAL {
 public:
    explicit FollowMessageTerminator(bool flag);

    // NOTE:
    // This function is a trick to write return statement which has no return value.
    // ex. return FollowMessageTerminator(true) = ...;
    // it is to simplify the else clause in the macros
    void operator=(const AssertionResult& msg) const;

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(FollowMessageTerminator);
};

inline bool suppress_warning_(bool b) { return b; }
NLIB_VIS_PUBLIC AssertionResult BooleanFailMessage(const char* file, int line, const char* expr,
                                                   const char* actual, const char* expected);
#define NLIB_TESTING_BOOLEAN_IMPL_(expr, expected, actual, iffail) \
    switch (0)                                                     \
    case 0:                                                        \
    default:                                                       \
        if (::nlib_ns::testing::suppress_warning_(expr))           \
            ;                                                      \
        else                                                       \
        iffail =                                                   \
            ::nlib_ns::testing::BooleanFailMessage(__FILE__, __LINE__, #expr, #actual, #expected)

NLIB_VIS_PUBLIC AssertionResult CompFailMessage(const char* file, int line, const char* expr1,
                                                const char* expr2, const char* val1,
                                                const char* val2, const char* op);
#define NLIB_TESTING_COMP_IMPL_(func_name, op)                                        \
template<class T1, class T2>                                                          \
AssertionResult Comp##func_name(const char* expr1, const char* expr2, const T1& val1, \
                                const T2& val2, const char* file, int line) {         \
        if (val1 op val2) {                                                           \
            return AssertionResult(true);                                             \
        } else {                                                                      \
            detail_func::DummyStream val1str, val2str;                                \
            val1str << val1;                                                          \
            val2str << val2;                                                          \
            return CompFailMessage(file, line, expr1, expr2, val1str.c_str(),         \
                                   val2str.c_str(), #op);                             \
        }                                                                             \
}                                                                                     \
NLIB_VIS_PUBLIC AssertionResult Comp##func_name(const char* expr1, const char* expr2, \
                                                int64_t val1, int64_t val2,           \
                                                const char* file, int line)

NLIB_TESTING_COMP_IMPL_(EQ, ==);
NLIB_TESTING_COMP_IMPL_(NE, !=);
NLIB_TESTING_COMP_IMPL_(LE, <=);
NLIB_TESTING_COMP_IMPL_(LT, <);
NLIB_TESTING_COMP_IMPL_(GE, >=);
NLIB_TESTING_COMP_IMPL_(GT, >);

#undef NLIB_TESTING_COMP_IMPL_

NLIB_VIS_PUBLIC AssertionResult CompStrEq(const char* expr1, const char* expr2, const char* val1,
                                          const char* val2, const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult CompStrNe(const char* expr1, const char* expr2, const char* val1,
                                          const char* val2, const char* file, int line);

NLIB_VIS_PUBLIC AssertionResult CompStrEq(const char* expr1, const char* expr2, const wchar_t* val1,
                                          const wchar_t* val2, const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult CompStrNe(const char* expr1, const char* expr2, const wchar_t* val1,
                                          const wchar_t* val2, const char* file, int line);

NLIB_VIS_PUBLIC AssertionResult CompStrCaseEq(const char* expr1, const char* expr2,
                                              const char* val1, const char* val2,
                                              const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult CompStrCaseNe(const char* expr1, const char* expr2,
                                              const char* val1, const char* val2,
                                              const char* file, int line);

NLIB_VIS_PUBLIC AssertionResult CompStrCaseEq(const char* expr1, const char* expr2,
                                              const wchar_t* val1, const wchar_t* val2,
                                              const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult CompStrCaseNe(const char* expr1, const char* expr2,
                                              const wchar_t* val1, const wchar_t* val2,
                                              const char* file, int line);

NLIB_VIS_PUBLIC AssertionResult CompFloatEq(const char* expr1, const char* expr2, float val1,
                                            float val2, const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult CompDoubleEq(const char* expr1, const char* expr2, double val1,
                                             double val2, const char* file, int line);
NLIB_VIS_PUBLIC AssertionResult NearDouble(double val1, double val2, double abs_error);

// NOTE: PRE05-C. Understand macro replacement
//                when concatenating tokens or performing stringification
#define NLIB_TESTING_CONCAT_TOKEN_(a, b) NLIB_TESTING_CONCAT_TOKEN_IMPL_(a, b)
#define NLIB_TESTING_CONCAT_TOKEN_IMPL_(a, b) a##b

#define NLIB_TESTING_FATAL_ return ::nlib_ns::testing::FollowMessageTerminator(true)
#define NLIB_TESTING_NONFATAL_ ::nlib_ns::testing::FollowMessageTerminator(false)

#define NLIB_TESTING_ASSERT_HELPER_(func, val1, val2)               \
    switch (0)                                                      \
    case 0:                                                         \
    default:                                                        \
        if (::nlib_ns::testing::AssertionResult ar =               \
                func(#val1, #val2, val1, val2, __FILE__, __LINE__)) \
            ;                                                       \
        else                                                        \
        NLIB_TESTING_FATAL_ = ar

#define NLIB_TESTING_EXPECT_HELPER_(func, val1, val2)               \
    switch (0)                                                      \
    case 0:                                                         \
    default:                                                        \
        if (::nlib_ns::testing::AssertionResult ar =               \
                func(#val1, #val2, val1, val2, __FILE__, __LINE__)) \
            ;                                                       \
        else                                                        \
        NLIB_TESTING_NONFATAL_ = ar

#define NLIB_TESTING_CLASS_NAME_(test_case, test_name) test_case##_##test_name##_Test

#define NLIB_TEST_(test_case, test_name, parent)                                          \
class NLIB_TESTING_CLASS_NAME_(test_case, test_name)                                      \
        : public parent {                                                                 \
public:                                                                                   \
        NLIB_TESTING_CLASS_NAME_(test_case, test_name)() {}                               \
private:                                                                                  \
        virtual void TestBody();                                                          \
        static ::nlib_ns::testing::TestInfo* info_;                                       \
        NLIB_DISALLOW_COPY_AND_ASSIGN(NLIB_TESTING_CLASS_NAME_(test_case, test_name));    \
};                                                                                        \
::nlib_ns::testing::TestInfo* NLIB_TESTING_CLASS_NAME_(test_case, test_name)::info_ =     \
        ::nlib_ns::testing::TestInfo::AddTestInfo<                                        \
            NLIB_TESTING_CLASS_NAME_(test_case, test_name)>(                              \
            #test_case, #test_name, -1, parent::SetUpTestCase, parent::TearDownTestCase); \
void NLIB_TESTING_CLASS_NAME_(test_case, test_name)::TestBody()

NLIB_VIS_PUBLIC AssertionResult& NearFailMessage(AssertionResult* ar, const char* file, int line,
                                                 const char* expr1, const char* expr2,
                                                 const char* abs_error_expr, double val1,
                                                 double val2, double abs_error);

NLIB_VIS_PUBLIC AssertionResult
    FloatLE(const char* expr1, const char* expr2, float val1, float val2);
NLIB_VIS_PUBLIC AssertionResult
    DoubleLE(const char* expr1, const char* expr2, double val1, double val2);
NLIB_VIS_PUBLIC AssertionResult AssertionSuccess();
NLIB_VIS_PUBLIC AssertionResult AssertionFailure();
NLIB_VIS_PUBLIC AssertionResult Message();

#define NLIB_TESTING_FLAG(flag_name) NLIB_FLAGS_##flag_name

NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_string(gtest_color);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_string(gtest_output);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_bool(gtest_break_on_failure);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_bool(gtest_also_run_disabled_tests);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_int32(gtest_repeat);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_string(gtest_filter);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_bool(gtest_shuffle);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_int32(gtest_random_seed);
NLIB_VIS_PUBLIC NLIB_FLAGS_DECLARE_bool(gtest_print_time);

}  // namespace testing
NLIB_NAMESPACE_END

NLIB_NAMESPACE_BEGIN
namespace testing {
#define TEST(test_case_name, specific_test_name) \
    NLIB_TEST_(test_case_name, specific_test_name, ::nlib_ns::testing::Test)
#define TEST_F(test_fixture_name, specific_test_name) \
    NLIB_TEST_(test_fixture_name, specific_test_name, test_fixture_name)

#define FRIEND_TEST(test_case, test_name) \
    friend class NLIB_TESTING_CLASS_NAME_(test_case, test_name)

#define RUN_ALL_TESTS() ::nlib_ns::testing::TestInfo::Run()

#define FAIL()                                                  \
    return ::nlib_ns::testing::FollowMessageTerminator(true) = \
               ::nlib_ns::testing::AssertionResult(false)
#define SUCCEED()

#define ASSERT_TRUE(expr) NLIB_TESTING_BOOLEAN_IMPL_(expr, true, false, NLIB_TESTING_FATAL_)
#define ASSERT_FALSE(expr) NLIB_TESTING_BOOLEAN_IMPL_(!(expr), false, true, NLIB_TESTING_FATAL_)
#define EXPECT_TRUE(expr) NLIB_TESTING_BOOLEAN_IMPL_(expr, true, false, NLIB_TESTING_NONFATAL_)
#define EXPECT_FALSE(expr) NLIB_TESTING_BOOLEAN_IMPL_(!(expr), false, true, NLIB_TESTING_NONFATAL_)

#define ASSERT_EQ(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompEQ, expected, actual)
#define ASSERT_NE(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompNE, expected, actual)
#define ASSERT_LE(val1, val2) NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompLE, val1, val2)
#define ASSERT_LT(val1, val2) NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompLT, val1, val2)
#define ASSERT_GE(val1, val2) NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompGE, val1, val2)
#define ASSERT_GT(val1, val2) NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompGT, val1, val2)

#define EXPECT_EQ(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompEQ, expected, actual)
#define EXPECT_NE(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompNE, expected, actual)
#define EXPECT_LE(val1, val2) NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompLE, val1, val2)
#define EXPECT_LT(val1, val2) NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompLT, val1, val2)
#define EXPECT_GE(val1, val2) NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompGE, val1, val2)
#define EXPECT_GT(val1, val2) NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompGT, val1, val2)

#define ASSERT_STREQ(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompStrEq, expected, actual)
#define ASSERT_STRNE(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompStrNe, expected, actual)

#define EXPECT_STREQ(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompStrEq, expected, actual)
#define EXPECT_STRNE(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompStrNe, expected, actual)

#define ASSERT_STRCASEEQ(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompStrCaseEq, expected, actual)
#define ASSERT_STRCASENE(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompStrCaseNe, expected, actual)

#define EXPECT_STRCASEEQ(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompStrCaseEq, expected, actual)
#define EXPECT_STRCASENE(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompStrCaseNe, expected, actual)

// NOTE: not implemented, it will be complicated because of locale
// ASSERT_STRCASEEQ
// ASSERT_STRCASENE

#define ASSERT_FLOAT_EQ(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompFloatEq, expected, actual)
#define ASSERT_DOUBLE_EQ(expected, actual) \
    NLIB_TESTING_ASSERT_HELPER_(::nlib_ns::testing::CompDoubleEq, expected, actual)

#define EXPECT_FLOAT_EQ(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompFloatEq, expected, actual)
#define EXPECT_DOUBLE_EQ(expected, actual) \
    NLIB_TESTING_EXPECT_HELPER_(::nlib_ns::testing::CompDoubleEq, expected, actual)

#define ASSERT_NEAR(val1, val2, abs_error)                              \
    switch (0)                                                          \
    case 0:                                                             \
    default:                                                            \
        if (::nlib_ns::testing::AssertionResult ar =                   \
                ::nlib_ns::testing::NearDouble(val1, val2, abs_error)) \
            ;                                                           \
        else                                                            \
        NLIB_TESTING_FATAL_ = ::nlib_ns::testing::NearFailMessage(     \
            &ar, __FILE__, __LINE__, #val1, #val2, #abs_error, val1, val2, abs_error)

#define EXPECT_NEAR(val1, val2, abs_error)                              \
    switch (0)                                                          \
    case 0:                                                             \
    default:                                                            \
        if (::nlib_ns::testing::AssertionResult ar =                   \
                ::nlib_ns::testing::NearDouble(val1, val2, abs_error)) \
            ;                                                           \
        else                                               /* NOLINT */ \
        NLIB_TESTING_NONFATAL_ = ::nlib_ns::testing::NearFailMessage(  \
            &ar, __FILE__, __LINE__, #val1, #val2, #abs_error, val1, val2, abs_error)

#define SCOPED_TRACE(msg)                                                                     \
    ::nlib_ns::testing::ScopedTrace NLIB_TESTING_CONCAT_TOKEN_(nlib_scopedtrace_, __LINE__)( \
        ::nlib_ns::testing::AssertionResult(true) << (msg), __FILE__, __LINE__)

#define ASSERT_NO_FATAL_FAILURE(statement)                                 \
    switch (0)                                                             \
    case 0:                                                                \
    default:                                                               \
        for (;;) {                                                         \
            statement;                                                     \
            break;                                                         \
        }                                                                  \
        if (!::nlib_ns::testing::UnitTest::GetInstance()                  \
                 ->current_test_info()                                     \
                 ->result()                                                \
                 ->HasFatalFaiulre())                                      \
            ;                                                              \
        else                                                 /* NOLINT */  \
        NLIB_TESTING_FATAL_ = ::nlib_ns::testing::AssertionResult(false)  \
                              << "Expected: " #statement                   \
                                 " doesn't generate new fatal failures.\n" \
                                 "  Actual: it does.\n"

#define EXPECT_NO_FATAL_FAILURE(statement)                                    \
    switch (0)                                                                \
    case 0:                                                                   \
    default:                                                                  \
        for (;;) {                                                            \
            statement;                                                        \
            break;                                                            \
        }                                                                     \
        if (!::nlib_ns::testing::UnitTest::GetInstance()                     \
                 ->current_test_info()                                        \
                 ->result()                                                   \
                 ->HasFatalFaiulre())                                         \
            ;                                                                 \
        else                                                                  \
        NLIB_TESTING_NONFATAL_ = ::nlib_ns::testing::AssertionResult(false)  \
                                 << "Expected: " #statement                   \
                                    " doesn't generate new fatal failures.\n" \
                                    "  Actual: it does.\n"

#define GTEST_FLAG(name) NLIB_FLAGS_gtest_##name

}  // namespace testing
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_testing_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_TESTING_NTEST_H_
