﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#ifndef SAMPLES_XML_RPC_XMLRPC_VALUE_H_  // NOLINT
#define SAMPLES_XML_RPC_XMLRPC_VALUE_H_

#include <map>
#include <string>
#include <vector>

#include "nn/nlib/exi/exi.h"

#define N(x) NLIB_EXI_LITERAL(x)
#define M(x) NLIB_EXI_UTF8(x)
typedef std::basic_string< ::nlib_ns::exi::ExiChar> StdString;

class XmlRpcArray;
class XmlRpcStruct;

class XmlRpcValue {
 public:
    enum ValueType {
        NONE = 0,
        INT,
        BOOLEAN,
        DOUBLE,
        STRING,
        NIL,
        // DATETIME and BASE64 are omitted
        // DATETIME,
        // BASE64
        ARRAY,
        STRUCT
    };

 private:
    union {
        int valueInt;
        bool valueBool;
        double valueDouble;
        StdString* valueString;
        XmlRpcArray* valueArray;
        XmlRpcStruct* valueStruct;
    } value_holder_;
    ValueType value_type_;
    char _[7];

 public:
    XmlRpcValue() {
        value_type_ = NONE;
        value_holder_.valueInt = 0;
    }
    ~XmlRpcValue() { this->Invalidate(); }
    void SetInt(const int& rhs) {
        this->Invalidate();
        value_type_ = INT;
        value_holder_.valueInt = rhs;
    }
    void SetBool(const bool& rhs) {
        this->Invalidate();
        value_type_ = BOOLEAN;
        value_holder_.valueBool = rhs;
    }
    void SetDouble(const double& rhs) {
        this->Invalidate();
        value_type_ = DOUBLE;
        value_holder_.valueDouble = rhs;
    }
    void SetString(const StdString& rhs) {
        this->Invalidate();
        value_type_ = STRING;
        value_holder_.valueString = new StdString(rhs);
    }
    void SetString(const nlib_ns::exi::ExiChar* rhs) {
        this->Invalidate();
        value_type_ = STRING;
        value_holder_.valueString = new StdString(rhs);
    }
    void SetArray();
    void SetStruct();

 public:
    ValueType GetValueType() const { return value_type_; }

    int* AsInt() { return (value_type_ == INT) ? &value_holder_.valueInt : NULL; }
    const int* AsInt() const { return (value_type_ == INT) ? &value_holder_.valueInt : NULL; }

    bool* AsBool() { return (value_type_ == BOOLEAN) ? &value_holder_.valueBool : NULL; }
    const bool* AsBool() const {
        return (value_type_ == BOOLEAN) ? &value_holder_.valueBool : NULL;
    }

    double* AsDouble() { return (value_type_ == DOUBLE) ? &value_holder_.valueDouble : NULL; }
    const double* AsDouble() const {
        return (value_type_ == DOUBLE) ? &value_holder_.valueDouble : NULL;
    }

    StdString* AsString() { return (value_type_ == STRING) ? value_holder_.valueString : NULL; }
    const StdString* AsString() const {
        return (value_type_ == STRING) ? value_holder_.valueString : NULL;
    }

    XmlRpcArray* AsArray() { return (value_type_ == ARRAY) ? value_holder_.valueArray : NULL; }
    const XmlRpcArray* AsArray() const {
        return (value_type_ == ARRAY) ? value_holder_.valueArray : NULL;
    }

    XmlRpcStruct* AsStruct() { return (value_type_ == STRUCT) ? value_holder_.valueStruct : NULL; }
    const XmlRpcStruct* AsStruct() const {
        return (value_type_ == STRUCT) ? value_holder_.valueStruct : NULL;
    }

    bool IsValid() const { return value_type_ != NONE; }
    bool IsNil() const { return value_type_ == NIL; }
    bool Read(nlib_ns::exi::XmlStreamReader* reader) {
        if (nlib_ns::exi::XmlStreamReader::START_ELEMENT != reader->Next()) return false;
        return this->Read_(reader);
    }
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcValue(const XmlRpcValue& rhs);             // forbidden
    XmlRpcValue& operator=(const XmlRpcValue& rhs);  // forbidden
    void Invalidate();
    bool Read_(nlib_ns::exi::XmlStreamReader* reader);

    friend class XmlRpcArray;
    friend class XmlRpcStruct;
};

class XmlRpcArray {
    typedef std::vector<XmlRpcValue*> ArrayValue;
    ArrayValue array_value_;

 public:
    XmlRpcValue* Append() {
        XmlRpcValue* p = new XmlRpcValue();
        array_value_.push_back(p);
        return p;
    }
    std::size_t Size() const { return array_value_.size(); }
    XmlRpcValue* Get(std::size_t idx) {
        return idx < array_value_.size() ? array_value_[idx] : NULL;
    }
    const XmlRpcValue* Get(std::size_t idx) const {
        return idx < array_value_.size() ? array_value_[idx] : NULL;
    }
    XmlRpcValue* operator[](std::size_t idx) { return this->Get(idx); }
    const XmlRpcValue* operator[](std::size_t idx) const { return this->Get(idx); }
    bool Read(nlib_ns::exi::XmlStreamReader* reader);
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcArray() {}
    ~XmlRpcArray();
    XmlRpcArray(const XmlRpcArray& rhs);             // forbidden
    XmlRpcArray& operator=(const XmlRpcArray& rhs);  // forbidden

    friend class XmlRpcValue;
};

class XmlRpcStruct {
    typedef std::map<StdString, XmlRpcValue*> StructValue;
    StructValue struct_value_;

 public:
    XmlRpcValue& operator[](const StdString& rhs);
    XmlRpcValue* Insert(const StdString& rhs) { return &operator[](rhs); }
    XmlRpcValue* Get(const StdString& rhs) const {
        StructValue::const_iterator it = struct_value_.find(rhs);
        return it != struct_value_.end() ? it->second : NULL;
    }
    bool HasMember(const StdString& rhs) const {
        return struct_value_.end() != struct_value_.find(rhs);
    }
    bool Read(nlib_ns::exi::XmlStreamReader* reader);
    void Write(nlib_ns::exi::XmlStreamWriter* writer) const;

 private:
    XmlRpcStruct() {}
    ~XmlRpcStruct();
    XmlRpcStruct(const XmlRpcStruct& rhs);             // forbidden
    XmlRpcStruct& operator=(const XmlRpcStruct& rhs);  // forbidden

    friend class XmlRpcValue;
};

#endif  // SAMPLES_XML_RPC_XMLRPC_VALUE_H_  // NOLINT
